"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable class-methods-use-this, no-underscore-dangle, no-use-before-define */
const crypto_1 = require("crypto");
const cosmosdb_query_1 = __importDefault(require("@zeit/cosmosdb-query"));
const lru_cache_1 = __importDefault(require("lru-cache"));
const get_value_1 = __importDefault(require("../utils/get-value"));
function ts() {
    return Math.floor(Date.now() / 1e3);
}
class Items {
    constructor(parent, partitionKeyPath, uniqueKeyPaths = []) {
        this._parent = parent;
        this._data = new Map();
        this._index = new Map();
        this._ridCount = 0;
        this._queryCache = new lru_cache_1.default({ max: 100 });
        this._partitionKeyPath = partitionKeyPath;
        this._uniqueKeyPaths = uniqueKeyPaths.concat("/id");
    }
    create(data) {
        const partition = this._getPartition(data);
        this._uniqueKeyPaths.forEach(keyPath => {
            const keyPathValue = (0, get_value_1.default)(keyPath.slice(1).split("/"), data);
            if (this._item(keyPath, keyPathValue, partition).read()) {
                const err = new Error(`Resource with specified ${keyPath.slice(1)} already exists.`);
                // @ts-ignore
                err.conflict = true;
                throw err;
            }
        });
        this._ridCount += 1;
        const _rid = this._rid(this._ridCount);
        const _data = {
            ...data,
            id: data.id,
            _etag: (0, crypto_1.randomUUID)(),
            _rid,
            _self: this._self(_rid),
            _ts: ts()
        };
        return this._set(_data);
    }
    delete(idOrRid, partition) {
        const data = this._item("/id", idOrRid, partition).read();
        if (!data) {
            throw new Error("does not exist");
        }
        this._uniqueKeyPaths.forEach(keyPath => {
            const keyPathValue = (0, get_value_1.default)(keyPath.slice(1).split("/"), data);
            const id = this._getId(keyPath, keyPathValue, this._getPartition(data));
            this._index.delete(id);
        });
        this._data.delete(data._rid);
    }
    query(params, { maxItemCount, continuation }) {
        if (!this._parent.read())
            return null;
        const data = [...this._data.values()].map(item => item.read());
        const udf = this._userDefinedFunctions();
        const parent = this._parent.read();
        const compositeIndexes = ((parent.indexingPolicy
            ? parent.indexingPolicy.compositeIndexes
            : null) || []);
        const q = this._getQuery(params.query);
        const options = {
            parameters: params.parameters,
            udf,
            maxItemCount,
            continuation,
            compositeIndexes
        };
        try {
            return q.exec(data, options);
        }
        catch (err) {
            // @ts-ignore
            err.badRequest = true;
            throw err;
        }
    }
    read({ maxItemCount, continuation } = {}) {
        if (!this._parent.read())
            return null;
        const data = [...this._data.values()].map(item => item.read());
        return this._getQuery("SELECT * FROM c").exec(data, {
            maxItemCount,
            continuation
        });
    }
    replace(data, original) {
        if (this._getPartition(data) !== this._getPartition(original)) {
            const err = new Error(`replacing partition key "${this._getPartitionKeyPath().join(".")}" is not allowed`);
            // @ts-ignore
            err.badRequest = true;
            throw err;
        }
        const partition = this._getPartition(data);
        const oldData = this._item("/id", data.id, partition).read();
        if (!oldData || data.id !== oldData.id) {
            const err = new Error("replacing id is not allowed");
            // @ts-ignore
            err.badRequest = true;
            throw err;
        }
        this._uniqueKeyPaths.forEach(keyPath => {
            // No need to check /id here since we explicitly checked it earlier
            if (keyPath === "/id") {
                return;
            }
            const keys = keyPath.slice(1).split("/");
            const newKeyPathValue = (0, get_value_1.default)(keys, data);
            const originalKeyPathValue = (0, get_value_1.default)(keys, original);
            if (newKeyPathValue === originalKeyPathValue) {
                return;
            }
            if (this._item(keyPath, newKeyPathValue, partition).read()) {
                const err = new Error(`Resource with specified ${keyPath.slice(1)} already exists.`);
                // @ts-ignore
                err.conflict = true;
                throw err;
            }
        });
        const { _rid, _self } = oldData;
        const _data = {
            ...data,
            id: data.id,
            _etag: (0, crypto_1.randomUUID)(),
            _rid,
            _self,
            _ts: ts()
        };
        this.delete(oldData.id, partition);
        return this._set(_data);
    }
    upsert(data) {
        const partition = this._getPartition(data);
        const oldData = this._item("/id", data.id, partition).read();
        return oldData ? this.replace(data, oldData) : this.create(data);
    }
    _item(keyPath, keyPathValue, partition) {
        const rid = this._toRid(keyPath, keyPathValue, partition);
        return this._data.get(rid) || this._newItem();
    }
    // eslint-disable-next-line no-unused-vars
    _newItem(data) {
        throw new Error("Not implemented");
    }
    _rid(id) {
        throw new Error("Not implemented");
    }
    // eslint-disable-next-line no-unused-vars
    _self(rid) {
        throw new Error("Not implemented");
    }
    _set(data) {
        this._uniqueKeyPaths.forEach(keyPath => {
            const keyPathValue = (0, get_value_1.default)(keyPath.slice(1).split("/"), data);
            const id = this._getId(keyPath, keyPathValue, this._getPartition(data));
            this._index.set(id, data._rid);
        });
        this._data.set(data._rid, this._newItem(data));
        return data;
    }
    _toRid(keyPath, keyPathValueOrRid, partition) {
        const id = this._getId(keyPath, keyPathValueOrRid, partition);
        return this._index.get(id) || keyPathValueOrRid;
    }
    _userDefinedFunctions() {
        return null;
    }
    _getQuery(sql) {
        let q = this._queryCache.get(sql);
        if (!q) {
            q = (0, cosmosdb_query_1.default)(sql);
            this._queryCache.set(sql, q);
        }
        return q;
    }
    _getId(keyPath, id, partition) {
        return [partition || id, keyPath, id].join(":");
    }
    _getPartition(data) {
        return (0, get_value_1.default)(this._getPartitionKeyPath(), data);
    }
    _getPartitionKeyPath() {
        const [firstPath] = this._partitionKeyPath;
        if (!firstPath) {
            return ["id"]; // Default/fallback to `id` as partition key for now
        }
        const path = firstPath.slice(1).split("/");
        if (path.length === 1 && path[0] === "_partitionKey") {
            return ["id"];
        }
        return path;
    }
}
exports.default = Items;
