"use strict";
// Derives from:
// https://github.com/Azure/azure-cosmos-js/blob/c1be5cc0fcc80670ae6c2a6b5f133530e7889db9/src/common/resourceId.ts
// https://github.com/Azure/azure-cosmosdb-java/blob/v2.6.0/commons/src/main/java/com/microsoft/azure/cosmosdb/internal/ResourceId.java
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EMPTY = void 0;
/* eslint-disable no-bitwise, no-plusplus */
const big_integer_1 = __importDefault(require("big-integer"));
const int64_buffer_1 = require("int64-buffer");
exports.EMPTY = "0"; // TODO: This is kinda hacky
class ResourceId {
    constructor() {
        this.offer = exports.EMPTY;
        this.database = exports.EMPTY;
        this.documentCollection = exports.EMPTY;
        this.storedProcedure = exports.EMPTY;
        this.trigger = exports.EMPTY;
        this.userDefinedFunction = exports.EMPTY;
        this.document = exports.EMPTY;
        this.partitionKeyRange = exports.EMPTY;
        this.user = exports.EMPTY;
        this.conflict = exports.EMPTY;
        this.permission = exports.EMPTY;
        this.attachment = exports.EMPTY;
    }
    static parse(id) {
        const pair = ResourceId.tryParse(id);
        if (!pair[0]) {
            throw new Error(`invalid resource id ${id}`);
        }
        return pair[1];
    }
    static newDatabaseId(dbId) {
        const resourceId = new ResourceId();
        resourceId.database = dbId;
        return resourceId;
    }
    static newDocumentCollectionId(databaseId, collectionId) {
        const dbId = ResourceId.parse(databaseId);
        const collectionResourceId = new ResourceId();
        collectionResourceId.database = dbId.database;
        collectionResourceId.documentCollection = collectionId;
        return collectionResourceId;
    }
    static newUserId(databaseId, userId) {
        const dbId = ResourceId.parse(databaseId);
        const userResourceId = new ResourceId();
        userResourceId.database = dbId.database;
        userResourceId.user = userId;
        return userResourceId;
    }
    static newPermissionId(userId, permissionId) {
        const usrId = ResourceId.parse(userId);
        const permissionResourceId = new ResourceId();
        permissionResourceId.database = usrId.database;
        permissionResourceId.user = usrId.user;
        permissionResourceId.permission = permissionId;
        return permissionResourceId;
    }
    static newAttachmentId(documentId, attachmentId) {
        const docId = ResourceId.parse(documentId);
        const attachmentResourceId = new ResourceId();
        attachmentResourceId.database = docId.database;
        attachmentResourceId.documentCollection = docId.documentCollection;
        attachmentResourceId.document = docId.document;
        attachmentResourceId.attachment = attachmentId;
        return attachmentResourceId;
    }
    static tryParse(id) {
        if (!id) {
            return [false, undefined];
        }
        const pair = ResourceId.verify(id);
        if (!pair[0]) {
            return [false, undefined];
        }
        const buffer = pair[1];
        const intArray = new Int8Array(buffer);
        if (buffer.length % 4 !== 0 && buffer.length !== ResourceId.OfferIdLength) {
            return [false, undefined];
        }
        const rid = new ResourceId();
        // if length < 4 bytes, the resource is an offer
        if (buffer.length === ResourceId.OfferIdLength) {
            let offer = 0;
            for (let index = 0; index < ResourceId.OfferIdLength; index++) {
                offer |= intArray[index] << (index * 8);
            }
            rid.offer = offer.toString();
            return [true, rid];
        }
        // first 4 bytes represent the database
        if (buffer.length >= 4) {
            rid.database = buffer.readIntBE(0, 4).toString();
        }
        if (buffer.length >= 8) {
            const isCollection = (intArray[4] & 128) > 0;
            if (isCollection) {
                // 5th - 8th bytes represents the collection
                rid.documentCollection = buffer.readIntBE(4, 4).toString();
                if (buffer.length >= 16) {
                    // 9th - 15th bytes represent one of document, trigger, sproc, udf, conflict, pkrange
                    const subCollectionResource = ResourceId.bigNumberReadIntBE(buffer, 8, 8).toString();
                    if (intArray[15] >> 4 === ResourceId.DocumentByte) {
                        rid.document = subCollectionResource;
                        // 16th - 20th bytes represent the attachment
                        if (buffer.length === 20) {
                            rid.attachment = buffer.readIntBE(16, 4).toString();
                        }
                    }
                    else if (Math.abs(intArray[15] >> 4) === ResourceId.StoredProcedureByte) {
                        rid.storedProcedure = subCollectionResource;
                    }
                    else if (intArray[15] >> 4 === ResourceId.TriggerByte) {
                        rid.trigger = subCollectionResource;
                    }
                    else if (intArray[15] >> 4 === ResourceId.UserDefinedFunctionByte) {
                        rid.userDefinedFunction = subCollectionResource;
                    }
                    else if (intArray[15] >> 4 === ResourceId.ConflictByte) {
                        rid.conflict = subCollectionResource;
                    }
                    else if (intArray[15] >> 4 === ResourceId.PartitionKeyRangeByte) {
                        rid.partitionKeyRange = subCollectionResource;
                    }
                    else {
                        return [false, rid];
                    }
                }
                else if (buffer.length !== 8) {
                    return [false, rid];
                }
            }
            else {
                // 5th - 8th bytes represents the user
                rid.user = buffer.readIntBE(4, 4).toString();
                // 9th - 15th bytes represent the permission
                if (buffer.length === 16) {
                    rid.permission = ResourceId.bigNumberReadIntBE(buffer, 8, 8).toString();
                }
                else if (buffer.length !== 8) {
                    return [false, rid];
                }
            }
        }
        return [true, rid];
    }
    static verify(id) {
        if (!id) {
            throw new Error(`invalid resource id ${id}`);
        }
        let buffer = ResourceId.fromBase64String(id);
        if (!buffer || buffer.length > ResourceId.Length) {
            buffer = undefined;
            return [false, buffer];
        }
        return [true, buffer];
    }
    static verifyBool(id) {
        return ResourceId.verify(id)[0];
    }
    static fromBase64String(s) {
        return Buffer.from(s.replace("-", "/"), "base64");
    }
    static toBase64String(buffer) {
        return buffer.toString("base64");
    }
    static bigNumberReadIntBE(buffer, offset, byteLength) {
        // eslint-disable-next-line no-param-reassign
        offset >>>= 0;
        // eslint-disable-next-line no-param-reassign
        byteLength >>>= 0;
        let i = byteLength;
        let mul = (0, big_integer_1.default)("1");
        let val = (0, big_integer_1.default)(buffer[offset + --i]);
        while (i > 0 && mul) {
            const temp = (0, big_integer_1.default)(buffer[offset + --i]);
            mul = mul.times(0x100);
            val = val.plus(temp.times(mul));
        }
        mul = mul.times(0x80);
        if (val.greater(mul)) {
            const subtrahend = (0, big_integer_1.default)(2);
            val = val.minus(subtrahend.pow(8 * byteLength));
        }
        return val;
    }
    isDatabaseId() {
        return (this.database !== exports.EMPTY &&
            (this.documentCollection === exports.EMPTY && this.user === exports.EMPTY));
    }
    getDatabaseId() {
        const rid = new ResourceId();
        rid.database = this.database;
        return rid;
    }
    getDocumentCollectionId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        return rid;
    }
    getUniqueDocumentCollectionId() {
        const db = (0, big_integer_1.default)(this.database);
        const coll = (0, big_integer_1.default)(this.documentCollection);
        return db
            .shiftLeft(32)
            .or(coll)
            .toString();
    }
    getStoredProcedureId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.storedProcedure = this.storedProcedure;
        return rid;
    }
    getTriggerId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.trigger = this.trigger;
        return rid;
    }
    getUserDefinedFunctionId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.userDefinedFunction = this.userDefinedFunction;
        return rid;
    }
    getConflictId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.conflict = this.conflict;
        return rid;
    }
    getDocumentId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.document = this.document;
        return rid;
    }
    getPartitonKeyRangeId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.partitionKeyRange = this.partitionKeyRange;
        return rid;
    }
    getUserId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.user = this.user;
        return rid;
    }
    getPermissionId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.user = this.user;
        rid.permission = this.permission;
        return rid;
    }
    getAttachmentId() {
        const rid = new ResourceId();
        rid.database = this.database;
        rid.documentCollection = this.documentCollection;
        rid.document = this.document;
        rid.attachment = this.attachment;
        return rid;
    }
    getOfferId() {
        const rid = new ResourceId();
        rid.offer = this.offer;
        return rid;
    }
    getValue() {
        let len = 0;
        if (this.offer !== exports.EMPTY) {
            len += ResourceId.OfferIdLength;
        }
        else if (this.database !== exports.EMPTY) {
            len += 4;
        }
        if (this.documentCollection !== exports.EMPTY || this.user !== exports.EMPTY) {
            len += 4;
        }
        if (this.document !== exports.EMPTY ||
            this.permission !== exports.EMPTY ||
            this.storedProcedure !== exports.EMPTY ||
            this.trigger !== exports.EMPTY ||
            this.userDefinedFunction !== exports.EMPTY ||
            this.conflict !== exports.EMPTY ||
            this.partitionKeyRange !== exports.EMPTY) {
            len += 8;
        }
        if (this.attachment !== exports.EMPTY) {
            len += 4;
        }
        const buffer = Buffer.alloc(len);
        buffer.fill(0);
        if (this.offer !== exports.EMPTY) {
            buffer.writeIntLE(Number(this.offer), 0, ResourceId.OfferIdLength);
        }
        else if (this.database !== exports.EMPTY) {
            buffer.writeIntBE(Number(this.database), 0, 4);
        }
        if (this.documentCollection !== exports.EMPTY) {
            buffer.writeIntBE(Number(this.documentCollection), 4, 4);
        }
        else if (this.user !== exports.EMPTY) {
            buffer.writeIntBE(Number(this.user), 4, 4);
        }
        let big;
        if (this.storedProcedure !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.storedProcedure);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.trigger !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.trigger);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.userDefinedFunction !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.userDefinedFunction);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.conflict !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.conflict);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.document !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.document);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.permission !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.permission);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        else if (this.partitionKeyRange !== exports.EMPTY) {
            big = new int64_buffer_1.Int64BE(this.partitionKeyRange);
            big.toBuffer().copy(buffer, 8, 0, 8);
        }
        if (this.attachment !== exports.EMPTY) {
            buffer.writeIntBE(Number(this.attachment), 16, 4);
        }
        return buffer;
    }
    toString() {
        return ResourceId.toBase64String(this.getValue());
    }
}
exports.default = ResourceId;
ResourceId.Length = 20;
ResourceId.OfferIdLength = 3;
ResourceId.DocumentByte = 0;
ResourceId.StoredProcedureByte = 8;
ResourceId.TriggerByte = 7;
ResourceId.UserDefinedFunctionByte = 6;
ResourceId.ConflictByte = 4;
ResourceId.PartitionKeyRangeByte = 5;
