"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.upsertOperation = exports.replaceOperation = exports.readOperation = exports.deleteOperation = exports.createOperation = void 0;
const get_partition_from_header_1 = require("../utils/get-partition-from-header");
function createOperation(collection, input) {
    if (!input.resourceBody || !input.resourceBody.id) {
        return {
            statusCode: 400,
            body: { message: "missing id" }
        };
    }
    let body;
    try {
        body = collection.documents.create(input.resourceBody);
    }
    catch (err) {
        if (err.conflict) {
            return {
                statusCode: 409,
                body: { code: "Conflict", message: err.message }
            };
        }
        throw err;
    }
    return {
        statusCode: 201,
        body
    };
}
exports.createOperation = createOperation;
function deleteOperation(collection, input) {
    const partitionKey = (0, get_partition_from_header_1.parsePartitionKey)(input.partitionKey, input.id);
    const data = collection.document(input.id, partitionKey).read();
    if (!data) {
        return {
            statusCode: 404,
            body: {}
        };
    }
    if (input.ifMatch && input.ifMatch !== data._etag) {
        return {
            statusCode: 412,
            body: {
                code: "PreconditionFailed",
                message: "Operation cannot be performed because one of the specified precondition is not met."
            }
        };
    }
    collection.documents.delete(input.id, partitionKey);
    return { statusCode: 204 };
}
exports.deleteOperation = deleteOperation;
function readOperation(collection, input) {
    const partitionKey = (0, get_partition_from_header_1.parsePartitionKey)(input.partitionKey, input.id);
    const data = collection.document(input.id, partitionKey).read();
    if (!data) {
        return {
            statusCode: 404,
            body: {
                code: "NotFound",
                message: "Entity with the specified id does not exist in the system.,"
            }
        };
    }
    return {
        statusCode: 200,
        body: data
    };
}
exports.readOperation = readOperation;
function replaceOperation(collection, input) {
    if (!input.resourceBody || !input.resourceBody.id) {
        return {
            statusCode: 400,
            body: { message: "missing id" }
        };
    }
    const { id } = input.resourceBody;
    const partitionKey = (0, get_partition_from_header_1.parsePartitionKey)(input.partitionKey, id);
    const data = collection.document(id, partitionKey).read();
    if (!data) {
        return {
            statusCode: 404,
            body: {}
        };
    }
    if (input.ifMatch && input.ifMatch !== data._etag) {
        return {
            statusCode: 412,
            body: {
                code: "PreconditionFailed",
                message: "Operation cannot be performed because one of the specified precondition is not met."
            }
        };
    }
    let body;
    try {
        body = collection.documents.replace(input.resourceBody, data);
    }
    catch (error) {
        if (error.badRequest) {
            return {
                statusCode: 400,
                body: {
                    code: "BadRequest",
                    message: error.message
                }
            };
        }
        if (error.conflict) {
            return {
                statusCode: 409,
                body: {
                    code: "Conflict",
                    message: error.message
                }
            };
        }
        throw error;
    }
    return { statusCode: 200, body };
}
exports.replaceOperation = replaceOperation;
function upsertOperation(collection, input) {
    if (!input.resourceBody || !input.resourceBody.id) {
        return {
            statusCode: 400,
            body: { message: "missing id" }
        };
    }
    const { id } = input.resourceBody;
    const partitionKey = (0, get_partition_from_header_1.parsePartitionKey)(input.partitionKey, id);
    const data = collection.document(id, partitionKey).read();
    if (input.ifMatch) {
        if (data && input.ifMatch !== data._etag) {
            return {
                statusCode: 412,
                body: {
                    code: "PreconditionFailed",
                    message: "Operation cannot be performed because one of the specified precondition is not met."
                }
            };
        }
    }
    const body = collection.documents.upsert(input.resourceBody);
    return {
        statusCode: data ? 200 : 201,
        body
    };
}
exports.upsertOperation = upsertOperation;
